#pragma once
#include "typemap/numeric.h"
#include "typemap/string.h"
#include "typemap/svapi.h"
#include "typemap/object.h"
#include "typemap/containers.h"
#include "typemap/auto.h"
#include "typemap/optional.h"
#include "typemap/error.h"
